#!BPY
"""
Name: 'Center of Mass'
Blender: 242
Group: 'Object'
Tooltip: 'Computes center of mass with four methods'
"""
__author__ = ["Zdislav"]
__url__ = ("blender", "blenderartists")
__version__ = "0"
__bpydoc__ = """\

Rules for calculs:

- NON-MANIFOLD MESHES (Meshes that contain edges shared by more than 2 faces) are not computed (can be changed below)
- WIRE MESHES (Meshes with no faces, only edges) will have the sum of their edges lengths computed and appear under LENGTH
- OPEN MESHES (Meshes that contain edges that are only shared by one face), will have the sume of the areas of their faces 
              computed and appear under AREA
- CLOSED MESHES (Meshes where all edges are shared by exactly two faces) will have their volume computed an appear under VOLUME

Important notes:

- Only meshes will be calculated
- Meshes made of several separated parts won't have their volume calculated correctly
- Some options can be changed in the first lines of the script
"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****

import Blender
from Blender import NMesh
from Blender.Mathutils import Vector, TriangleArea

# --------------------------------------------------------------------

# CONSTANTS THAT CAN BE CHANGED 

COMPUTEOPENMESHES = False						# if False: only calculate area of open meshes,
												# if True: calculate volume of open meshes (warning, strange results)

COMPUTEMANIFOLD = False							# if False: non-manifold meshes are discarded


# --------------------------------------------------------------------

def triangulateNMesh(nm):					# function copied from the Blending into Python Wikibook - see mediawiki.blender.org
        '''
        Converts the meshes faces to tris, modifies the mesh in place.
        '''
        
        #============================================================================#
        # Returns a new face that has the same properties as the origional face      #
        # but with no verts                                                          #
        #============================================================================#
        def copyFace(face):
                newFace = NMesh.Face()
                # Copy some generic properties
                newFace.mode = face.mode
                if face.image != None:
                        newFace.image = face.image
                newFace.flag = face.flag
                newFace.mat = face.mat
                newFace.smooth = face.smooth
                return newFace
        
        # 2 List comprehensions are a lot faster then 1 for loop.
        tris = [f for f in nm.faces if len(f) == 3]
        quads = [f for f in nm.faces if len(f) == 4]
        
        
        if quads: # Mesh may have no quads.
                has_uv = quads[0].uv 
                has_vcol = quads[0].col
                for quadFace in quads:
                        
                        # Triangulate along the shortest edge
                        if (quadFace.v[0].co - quadFace.v[2].co).length < (quadFace.v[1].co - quadFace.v[3].co).length:
                                # Method 1
                                triA = 0,1,2
                                triB = 0,2,3
                        else:
                                # Method 2
                                triA = 0,1,3
                                triB = 1,2,3
                                
                        for tri1, tri2, tri3 in (triA, triB):
                                newFace = copyFace(quadFace)
                                newFace.v = [quadFace.v[tri1], quadFace.v[tri2], quadFace.v[tri3]]
                                if has_uv: newFace.uv = [quadFace.uv[tri1], quadFace.uv[tri2], quadFace.uv[tri3]]
                                if has_vcol: newFace.col = [quadFace.col[tri1], quadFace.col[tri2], quadFace.col[tri3]]
                                
                                nm.addEdge(quadFace.v[tri1], quadFace.v[tri3]) # Add an edge where the 2 tris are devided.
                                tris.append(newFace)
                
                nm.faces = tris
                

# --------------------------------------------------------------------               

def calculateVolume(mymesh):				# volume calculation
	triangulateNMesh(mymesh)
	vtot = 0
	for f in mymesh.faces:
		fzn = f.normal[2]
		x1,y1,z1 = f.v[0]
		x2,y2,z2 = f.v[1]
		x3,y3,z3 = f.v[2]
		volume = ((z1+z2+z3)/6.0)*abs((x1*(y3-y2))+(x2*(y1-y3))+(x3*(y2-y1)))
		if fzn < 0:
			vtot -= volume
		elif fzn > 0:
			vtot += volume
	return vtot

# --------------------------------------------------------------------
	
def calculateArea(mymesh):					# area calculation
	triangulateNMesh(mymesh)
	atot = 0
	for f in mymesh.faces:
		a = TriangleArea(f.v[0].co, f.v[1].co, f.v[2].co)
		atot = atot + a
	return atot

# --------------------------------------------------------------------
					
def calculateLength(mymesh):				# length calculation
	ltot = 0
	for e in mymesh.edges:
		ltot = ltot + (e.v1.co-e.v2.co).length
	return ltot

# --------------------------------------------------------------------
		
def CoMFromVerts(mymesh):
	num = len(mymesh.verts)
	com = Vector((0.0,0.0,0.0))
	for vert in mymesh.verts:
		com += vert.co
	return com/num

# --------------------------------------------------------------------

def CoMFromEdges(mymesh):
	length = 0
	com = Vector((0.0,0.0,0.0))
	for e in mymesh.edges:
		elength = (e.v1.co-e.v2.co).length
		com += (e.v1.co+e.v2.co) / 2.0 * elength
		length += elength
	return com/length

# --------------------------------------------------------------------

def CoMFromFaces(mymesh):
	triangulateNMesh(mymesh)
	area = 0
	com = Vector((0.0,0.0,0.0))
	for f in mymesh.faces:
		farea = TriangleArea(f.v[0].co,f.v[1].co,f.v[2].co)
		com += (f.v[0].co+f.v[1].co+f.v[2].co)/3.0 * farea
		area += farea
	return com/area

# --------------------------------------------------------------------

def CoMFromVolume(mymesh):
	triangulateNMesh(mymesh)
	com = Vector((0.0,0.0,0.0))
	vtot = 0
	for f in mymesh.faces:
		fzn = f.normal[2]
		x1,y1,z1 = f.v[0]
		x2,y2,z2 = f.v[1]
		x3,y3,z3 = f.v[2]
		pa = abs((x1*(y3-y2))+(x2*(y1-y3))+(x3*(y2-y1))) / 2.0
		volume = ((z1+z2+z3)/3.0) * pa
		z1,z2,z3 = abs(z1),abs(z2),abs(z3)
		sumz = z1+z2+z3
		cx = (x1*z1 + x2*z2 + x3*z3) / sumz
		cy = (y1*z1 + y2*z2 + y3*z3) / sumz
		cz = sumz / 6.0
		c = Vector((cx,cy,cz))
		if fzn < 0:
			com -= volume * c
		elif fzn > 0:
			com += volume * c
		vtot += volume
	return com/vtot

# --------------------------------------------------------------------

def manifoldType(mymesh):
	for e in mymesh.edges:
		shared=0
		for f in mymesh.faces:
			for vf1 in f.v:
				if vf1 == e.v1:
					for vf2 in f.v:
						if vf2 == e.v2:
							shared = shared+1
		if (shared > 2):
			return "non-manifold"
		if (shared < 2):
			return "open"
	return "closed"

# --------------------------------------------------------------------


# main
		
selection = Blender.Object.GetSelected()
notmeshes = []
nonmanifs = []
if len(selection) == 0:
	print 'Error: No Objects Selected'
else:
	print "Name	Volume	Area	Length	Center of Mass x y z"
	for obj in selection:
		volume = area = length = 0
		if obj.getType() != 'Mesh':		# discard objects
			reason =  'Object ' + obj.name + ' is not a mesh, not computed.'
			notmeshes.append(obj.name)
		else:
			mymesh = NMesh.GetRawFromObject(obj.name)
			# Scale the mesh (including normals) so that it can be compared to other world objects.
			wmatrix = obj.getMatrix()
			mymesh.transform(wmatrix, 1)
			
			mymeshType = manifoldType(mymesh)
			if (mymeshType == "non-manifold"):
				if not COMPUTEMANIFOLD:
					reason =  'Object ' + obj.name + ' is non-manifold (some edges are shared by more than 2 faces), not computed.'
					nonmanifs.append(obj.name)
 			if mymesh.edges:
 				length = calculateLength(mymesh)
			if mymesh.faces:
	 			if mymeshType == "closed":
	 				volume = calculateVolume(mymesh)
	 			else:
	 				if COMPUTEOPENMESHES:
						volume = calculateVolume(mymesh)
				area = calculateArea(mymesh)
		print "%s	%.3f	%.3f	%.3f " % (obj.name, volume, area, length)
		comV = CoMFromVerts(mymesh)
		print "	verts	%6.2f %6.2f %6.2f " % (comV[0],comV[1],comV[2])
		comV = CoMFromEdges(mymesh)
		print "	edges	%6.2f %6.2f %6.2f " % (comV[0],comV[1],comV[2])
		comV = CoMFromFaces(mymesh)
		print "	faces	%6.2f %6.2f %6.2f " % (comV[0],comV[1],comV[2])
		comV = CoMFromVolume(mymesh)
		print "	volume 	%6.2f %6.2f %6.2f " % (comV[0],comV[1],comV[2])
	if notmeshes:
		print 'Object(s) ' + ', '.join(notmeshes) + ' are not a mesh, not computed.'
	if nonmanifs:
		print 'Object(s) ' + ', '.join(morematls) + ' are non-manifold (some edges are shared by more than 2 faces), not computed.'